#!/usr/bin/env python3

# Numpy
import numpy as np

# OpenCV
import cv2 as cv

# Ros libraries
import roslib
import rospy

# Ros Messages
from sensor_msgs.msg import CompressedImage
from cup_detection_msgs.msg import Cups
""" CONSTANTES """

# Debug
VERBOSE = False


# Teinte
h_red = 175 #350° 
h_green = 70 #140°
h_var = 5 #+|-10°

# Saturation
low_sat = 70 #~24%
high_sat = 255 #100%

# Valeur
low_bright = 50 #~20%
high_bright = 255 #100%

# Red color
low_red_1 = np.array([h_red-h_var, low_sat, low_bright])
high_red_1 = np.array([h_red+h_var, high_sat, high_bright])
low_red_2 = np.array([255, low_sat, low_bright])
high_red_2 = np.array([255, high_sat, high_bright])

# Green color
low_green = np.array([h_green-h_var, low_sat, low_bright])
high_green = np.array([h_green+h_var, high_sat, high_bright])

# Kernel
kernel = np.ones((21,21),np.uint8)

""" Fonctions """

def get_red(frame):
    hsv_frame = cv.cvtColor(frame, cv.COLOR_BGR2HSV)
    # red hsv
    red_mask_1 = cv.inRange(hsv_frame, low_red_1, high_red_1)
    red_mask_2 = cv.inRange(hsv_frame, low_red_2, high_red_2)
    red_mask = cv.bitwise_or(red_mask_1,red_mask_2)
    red_trf = cv.morphologyEx(red_mask, cv.MORPH_OPEN, kernel)
    contours,_ = cv.findContours(red_trf, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)

    return contours 

def get_green(frame):
    hsv_frame = cv.cvtColor(frame, cv.COLOR_BGR2HSV)
    # green hsv
    green_mask = cv.inRange(hsv_frame, low_green, high_green)
    green_trf = cv.morphologyEx(green_mask, cv.MORPH_CLOSE, kernel)
    contours,_ = cv.findContours(green_trf, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)
    
    return contours

def get_pose(x, y, w, h, width, height):
    new_x = 2*(x+w/2-width/2)/width
    new_y = 2*(y+h/2-height/2)/height
    return new_x, new_y
""" Class """

class raspicam_treatment:

    def __init__(self):
        '''Initialize ros subscriber and publisher'''
        # subscribed Topic
        self.subscriber = rospy.Subscriber("/raspicam_node/image/compressed",
            CompressedImage, self.callback, queue_size = 1)
        # publisher Topic
        self.publisher = rospy.Publisher("cup_detection", Cups, queue_size = 1)

    def callback(self, ros_data):
        '''Callback function of subscribed topic.
        Here images get converted and features detected'''

        #### direct conversion to CV ####
        np_arr = np.frombuffer(ros_data.data, np.uint8)
        frame = cv.imdecode(np_arr, cv.IMREAD_COLOR) # OpenCV >= 3.0:
        frame = cv.rotate(frame, cv.ROTATE_180)

        #### Feature detectors using CV ####
        frame_final = frame
        height, width = frame.shape[:2]
        red_contours = get_red(frame)
        green_contours = get_green(frame)

        #### Message Generation ####
        msg = Cups()
        msg.size = 0
        color = np.empty(shape=(0, 0), dtype=np.uint8)
        pose = np.empty(shape=(0, 0), dtype=np.single)

        for cnt in red_contours:
            (x, y, w, h) = cv.boundingRect(cnt)
            if max(w/h,h/w)>1 and  max(w/h,h/w)<1.9: 
                new_x, new_y = get_pose(x, y, w, h, width, height)
                msg.size=msg.size+1
                color.append(msg.RED)
                pose.append([new_x, new_y])
                # DEBUG
                if VERBOSE :
                    cv.putText(frame_final,'{0}, {1}'.format(new_x, new_y),(x+int(w/2),y+int(h/2)), cv.FONT_HERSHEY_SIMPLEX, 1,(255,255,255),2)

        for cnt in green_contours:
            (x, y, w, h) = cv.boundingRect(cnt)
            if max(w/h,h/w)>1 and  max(w/h,h/w)<1.9: 
                new_x, new_y = get_pose(x, y, w, h, width, height)
                msg.size=msg.size+1
                color.append(msg.GREEN)
                pose.append([new_x, new_y])
                # DEBUG
                if VERBOSE :
                    cv.putText(frame_final,'{0}, {1}'.format(new_x, new_y),(x+int(w/2),y+int(h/2)), cv.FONT_HERSHEY_SIMPLEX, 1,(255,255,255),2)
        # DEBUG
        if VERBOSE :
            cv.imshow("frame_final",frame_final)
        # Publish new image
        if msg.size>0:
            msg.color = color
            msg.pose = pose
            self.publisher.publish(msg)
""" Main """

def main():
    '''Initializes and cleanup ros node'''
    ic = raspicam_treatment()
    rospy.init_node('raspicam_treatment_node', anonymous=True)
    try:
        rospy.spin()
    except KeyboardInterrupt:
        print("Shutting down ROS Image feature detector module")
    cv.destroyAllWindows()

if __name__ == '__main__':
    main()

